import warnings
from datetime import datetime

from pymongo import ASCENDING, DESCENDING

from mongo_thingy import AsyncThingy, BaseThingy, Thingy
from mongo_thingy.cursor import AsyncCursor, BaseCursor, Cursor


class BaseRevisionCursor(BaseCursor):
    def __init__(self, *args, **kwargs):
        super(BaseRevisionCursor, self).__init__(*args, **kwargs)


class RevisionCursor(Cursor, BaseRevisionCursor):
    def __getitem__(self, index):
        if index < 0 and self.thingy:
            version = self.thingy.count_revisions()
            index = version + index
        return super(RevisionCursor, self).__getitem__(index)


class AsyncRevisionCursor(AsyncCursor, BaseRevisionCursor):
    pass


class BaseRevision(BaseThingy):
    """Revision of a document"""

    _collection_name = "revision"
    _cursor_cls = None

    @classmethod
    def from_thingy(cls, thingy, author=None, operation="update"):
        version = cls(
            document_id=thingy.id,
            document_type=type(thingy).__name__,
            operation=operation,
        )
        if operation != "delete":
            version.document = thingy.__dict__
        if author:
            version.author = author
        return version

    def _prev(self):
        return self.find_one(
            {"document_id": self.document_id, "document_type": self.document_type}
        )


BaseRevision.add_index([("document_id", DESCENDING), ("document_type", DESCENDING)])


class Revision(Thingy, BaseRevision):
    _cursor_cls = RevisionCursor

    def save(self):
        self.creation_date = datetime.utcnow()
        if not self._prev():
            self.operation = "create"
        return super(Revision, self).save()


class AsyncRevision(AsyncThingy, BaseRevision):
    _cursor_cls = AsyncRevisionCursor

    async def save(self):
        self.creation_date = datetime.utcnow()
        if not await self._prev():
            self.operation = "create"
        return await super(AsyncRevision, self).save()


class BaseVersioned:
    """Mixin to versionate changes in a collection"""

    _revisions_cls = None

    def get_revisions(self, **kwargs):
        filter = {"document_id": self.id, "document_type": type(self).__name__}
        filter.update(kwargs)

        cursor = self._revision_cls.find(filter)
        cursor.thingy = self
        return cursor.sort("_id", ASCENDING)

    def count_revisions(self, **kwargs):
        filter = {"document_id": self.id, "document_type": type(self).__name__}
        return self._revision_cls.count_documents(filter, **kwargs)


class Versioned(BaseVersioned):
    def is_versioned(self):
        return bool(self.count_revisions(limit=1))

    @property
    def version(self):
        warnings.warn(
            "version is deprecated. Use count_revisions() instead.", DeprecationWarning
        )
        return self.count_revisions()

    @property
    def versioned(self):
        warnings.warn(
            "versioned is deprecated. Use is_versioned() instead.", DeprecationWarning
        )
        return bool(self.count_revisions(limit=1))

    @property
    def revisions(self):
        warnings.warn(
            "revisions is deprecated. Use get_revisions() instead.", DeprecationWarning
        )
        return self.get_revisions()

    def revert(self):
        revisions = list(self.get_revisions().limit(3))
        try:
            self.__dict__ = revisions[-2].document
        except IndexError:
            self.__dict__ = {"_id": self.id}
        return self.save()

    def save(self, author=None, **kwargs):
        result = super(Versioned, self).save(**kwargs)
        version = self._revision_cls.from_thingy(self, author=author)
        version.save()
        return result

    def delete(self, author=None):
        result = super(Versioned, self).delete()
        version = self._revision_cls.from_thingy(
            self, author=author, operation="delete"
        )
        version.save()
        return result


class AsyncVersioned(BaseVersioned):
    async def is_versioned(self):
        return bool(await self.count_revisions(limit=1))

    async def revert(self):
        revisions = await self.get_revisions().limit(3).to_list(length=3)
        try:
            self.__dict__ = revisions[-2].document
        except IndexError:
            self.__dict__ = {"_id": self.id}
        return await self.save()

    async def save(self, author=None, **kwargs):
        result = await super(AsyncVersioned, self).save(**kwargs)
        version = self._revision_cls.from_thingy(self, author=author)
        await version.save()
        return result

    async def delete(self, author=None):
        result = await super(AsyncVersioned, self).delete()
        version = self._revision_cls.from_thingy(
            self, author=author, operation="delete"
        )
        await version.save()
        return result


__all__ = ["AsyncRevision", "AsyncVersioned", "Revision", "Versioned"]



import asyncio
import warnings
from collections.abc import Mapping

from pymongo import MongoClient, ReturnDocument
from pymongo.errors import ConfigurationError
from thingy import DatabaseThingy, classproperty, registry

from mongo_thingy.cursor import AsyncCursor, Cursor

try:
    from motor.motor_tornado import MotorClient
except ImportError:  # pragma: no cover
    MotorClient = None

try:
    from motor.motor_asyncio import AsyncIOMotorClient
except ImportError:  # pragma: no cover
    AsyncIOMotorClient = None


class ThingyList(list):
    def distinct(self, key):
        def __get_value(document):
            if hasattr(document, "__dict__"):
                document = document.__dict__
            return document.get(key)

        values = set(__get_value(result) for result in self)
        return list(values)


class BaseThingy(DatabaseThingy):
    """Represents a document in a collection"""

    _client = None
    _client_cls = None
    _collection = None
    _collection_name = None
    _cursor_cls = None
    _result_cls = ThingyList

    @classproperty
    def _table(cls):
        return cls._collection

    @classproperty
    def _table_name(cls):
        return cls._collection_name

    @classproperty
    def table_name(cls):
        return cls.collection_name

    @classproperty
    def collection(cls):
        return cls.get_collection()

    @classproperty
    def collection_name(cls):
        return cls.get_table_name()

    @classproperty
    def client(cls):
        return cls.get_client()

    @classmethod
    def _get_client(cls, database):
        return database.client

    @classmethod
    def _get_database(cls, collection, name):
        if collection is not None:
            return collection.database
        if cls._client and name:
            return cls._client[name]
        raise AttributeError("Undefined database.")

    @classmethod
    def _get_table(cls, database, table_name):
        return database[table_name]

    @classmethod
    def _get_database_name(cls, database):
        return database.name

    @classmethod
    def _get_table_name(cls, table):
        return table.name

    @classmethod
    def get_client(cls):
        if cls._client:
            return cls._client
        return cls._get_client(cls.database)

    @classmethod
    def get_collection(cls):
        return cls.get_table()

    @classmethod
    def add_index(cls, keys, **kwargs):
        kwargs.setdefault("background", True)
        if not hasattr(cls, "_indexes"):
            cls._indexes = []
        cls._indexes.append((keys, kwargs))

    @classmethod
    def count_documents(cls, filter=None, *args, **kwargs):
        if filter is None:
            filter = {}
        return cls.collection.count_documents(filter, *args, **kwargs)

    @classmethod
    def count(cls, filter=None, *args, **kwargs):
        warnings.warn(
            "count is deprecated. Use count_documents instead.", DeprecationWarning
        )
        return cls.count_documents(filter=filter, *args, **kwargs)

    @classmethod
    def connect(cls, *args, client_cls=None, database_name=None, **kwargs):
        if not client_cls:
            client_cls = cls._client_cls

        cls._client = client_cls(*args, **kwargs)
        try:
            cls._database = cls._client.get_database(database_name)
        except (ConfigurationError, TypeError):
            cls._database = cls._client["test"]

    @classmethod
    def disconnect(cls, *args, **kwargs):
        if cls._client:
            cls._client.close()
        cls._client = None
        cls._database = None

    @classmethod
    def distinct(cls, *args, **kwargs):
        return cls.collection.distinct(*args, **kwargs)

    @classmethod
    def find(cls, *args, view=None, **kwargs):
        delegate = cls.collection.find(*args, **kwargs)
        return cls._cursor_cls(delegate, thingy_cls=cls, view=view)

    @classmethod
    def find_one(cls, filter=None, *args, **kwargs):
        if filter is not None and not isinstance(filter, Mapping):
            filter = {"_id": filter}

        cursor = cls.find(filter, *args, **kwargs)
        return cursor.first()

    @property
    def id(self):
        return self.__dict__.get("id") or self._id

    @id.setter
    def id(self, value):
        if "id" in self.__dict__:
            self.__dict__["id"] = value
        else:
            self._id = value

    def delete(self):
        return self.get_collection().delete_one({"_id": self.id})


class Thingy(BaseThingy):
    _client_cls = MongoClient
    _cursor_cls = Cursor

    @classmethod
    def create_index(cls, keys, **kwargs):
        cls.add_index(keys, **kwargs)
        cls.collection.create_index(keys, **kwargs)

    @classmethod
    def create_indexes(cls):
        if hasattr(cls, "_indexes"):
            for keys, kwargs in cls._indexes:
                cls.collection.create_index(keys, **kwargs)

    @classmethod
    def find_one_and_replace(cls, *args, **kwargs):
        kwargs.setdefault("return_document", ReturnDocument.AFTER)
        result = cls.collection.find_one_and_replace(*args, **kwargs)
        if result is not None:
            return cls(result)

    def save(self, force_insert=False):
        data = self.__dict__
        if self.id is not None and not force_insert:
            filter = {"_id": self.id}
            self.get_collection().replace_one(filter, data, upsert=True)
        else:
            self.get_collection().insert_one(data)
        return self


class AsyncThingy(BaseThingy):
    _client_cls = MotorClient or AsyncIOMotorClient
    _cursor_cls = AsyncCursor

    @classmethod
    async def create_index(cls, keys, **kwargs):
        cls.add_index(keys, **kwargs)
        await cls.collection.create_index(keys, **kwargs)

    @classmethod
    async def create_indexes(cls):
        if hasattr(cls, "_indexes"):
            for keys, kwargs in cls._indexes:
                await cls.collection.create_index(keys, **kwargs)

    @classmethod
    async def find_one_and_replace(cls, *args, **kwargs):
        kwargs.setdefault("return_document", ReturnDocument.AFTER)
        result = await cls.collection.find_one_and_replace(*args, **kwargs)
        if result is not None:
            return cls(result)

    async def save(self, force_insert=False):
        data = self.__dict__
        if self.id is not None and not force_insert:
            filter = {"_id": self.id}
            await self.get_collection().replace_one(filter, data, upsert=True)
        else:
            await self.get_collection().insert_one(data)
        return self


def connect(*args, **kwargs):
    if AsyncThingy._client_cls is not None:
        AsyncThingy.connect(*args, **kwargs)
    Thingy.connect(*args, **kwargs)


def disconnect(*args, **kwargs):
    Thingy.disconnect(*args, **kwargs)
    AsyncThingy.disconnect(*args, **kwargs)


def create_indexes():
    """Create indexes registered on all :class:`Thingy`"""
    for cls in registry:
        if issubclass(cls, Thingy):
            cls.create_indexes()
        if issubclass(cls, AsyncThingy):
            coroutine = cls.create_indexes()
            asyncio.create_task(coroutine)


__all__ = ["AsyncThingy", "Thingy", "connect", "create_indexes"]



import functools


class _Proxy:
    def __init__(self, name):
        self.name = name

    def __call__(self, cursor):
        return getattr(cursor.delegate, self.name)


class _ChainingProxy(_Proxy):
    def __call__(self, cursor):
        method = getattr(cursor.delegate, self.name)

        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            method(*args, **kwargs)
            return cursor

        return wrapper


class _BindingProxy(_Proxy):
    def __call__(self, cursor):
        method = getattr(cursor.delegate, self.name)

        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            result = method(*args, **kwargs)
            return cursor.bind(result)

        return wrapper


class _AsyncBindingProxy(_Proxy):
    def __call__(self, cursor):
        method = getattr(cursor.delegate, self.name)

        @functools.wraps(method)
        async def wrapper(*args, **kwargs):
            result = await method(*args, **kwargs)
            if isinstance(result, list):
                return cursor.result_cls(cursor.bind(r) for r in result)
            return cursor.bind(result)

        return wrapper


class BaseCursor:
    distinct = _Proxy("distinct")
    explain = _Proxy("explain")
    limit = _ChainingProxy("limit")
    skip = _ChainingProxy("skip")
    sort = _ChainingProxy("sort")

    def __init__(self, delegate, thingy_cls=None, view=None):
        self.delegate = delegate
        self.thingy_cls = thingy_cls
        self.result_cls = getattr(thingy_cls, "_result_cls", list)

        if isinstance(view, str):
            view = self.get_view(view)

        self.thingy_view = view

    def __getattribute__(self, name):
        attribute = object.__getattribute__(self, name)
        if isinstance(attribute, _Proxy):
            return attribute(self)
        return attribute

    def bind(self, document):
        if not self.thingy_cls:
            return document
        thingy = self.thingy_cls(document)
        if self.thingy_view is not None:
            return self.thingy_view(thingy)
        return thingy

    def clone(self):
        delegate = self.delegate.clone()
        return self.__class__(
            delegate, thingy_cls=self.thingy_cls, view=self.thingy_view
        )

    def get_view(self, name):
        return self.thingy_cls._views[name]

    def view(self, name="defaults"):
        self.thingy_view = self.get_view(name)
        return self


class Cursor(BaseCursor):
    next = __next__ = _BindingProxy("__next__")

    def __getitem__(self, index):
        document = self.delegate.__getitem__(index)
        return self.bind(document)

    def to_list(self, length):
        self.limit(length)
        return self.result_cls(self)

    def delete(self):
        ids = self.distinct("_id")
        return self.thingy_cls.collection.delete_many({"_id": {"$in": ids}})

    def first(self):
        try:
            document = self.delegate.clone().limit(-1).__next__()
        except StopIteration:
            return None
        return self.bind(document)


class AsyncCursor(BaseCursor):
    to_list = _AsyncBindingProxy("to_list")
    next = __anext__ = _AsyncBindingProxy("__anext__")

    async def __aiter__(self):
        async for document in self.delegate:
            yield self.bind(document)

    async def delete(self):
        ids = await self.distinct("_id")
        return await self.thingy_cls.collection.delete_many({"_id": {"$in": ids}})

    async def first(self):
        try:
            document = await self.delegate.clone().limit(-1).__anext__()
        except StopAsyncIteration:
            return None
        return self.bind(document)


__all__ = ["AsyncCursor", "Cursor"]



from setuptools import setup


def get_description():
    with open("README.md") as file:
        return file.read()


setup(
    name="Mongo-Thingy",
    version="0.15.1",
    url="https://github.com/Refty/mongo-thingy",
    license="MIT",
    author="Guillaume Gelin",
    author_email="guillaume@refty.co",
    description=(
        "The most idiomatic and friendly-yet-powerful way to use " "MongoDB with Python"
    ),
    long_description=get_description(),
    long_description_content_type="text/markdown",
    packages=["mongo_thingy"],
    include_package_data=True,
    zip_safe=False,
    platforms="any",
    install_requires=["thingy>=0.9.0", "pymongo>=4"],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Database",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
